@@ -1,21 +1,22 @@
 use arloader::{
-    error::ArweaveError,
+    error::Error,
+    solana::{FLOOR, RATE, SOL_AR_BASE_URL},
     status::{OutputFormat, OutputHeader, Status, StatusCode},
     transaction::{Base64, FromStrs, Tag},
-    update_statuses_stream, upload_files_stream, Arweave, Methods as ArweaveMethods,
+    update_statuses_stream, upload_files_stream, upload_files_with_sol_stream, Arweave,
     WINSTONS_PER_AR,
 };
-
 use clap::{
     self, crate_description, crate_name, crate_version, App, AppSettings, Arg, SubCommand, Values,
 };
 use futures::StreamExt;
 use glob::glob;
 use num_traits::cast::ToPrimitive;
+use solana_sdk::signer::keypair;
 use std::{fmt::Display, path::PathBuf, str::FromStr};
 use url::Url;
 
-pub type CommandResult = Result<(), ArweaveError>;
+pub type CommandResult = Result<(), Error>;
 
 fn is_parsable_generic<U, T>(string: T) -> Result<(), String>
 where
@@ -95,6 +96,87 @@ fn get_status_code(output: &str) -> StatusCode {
     }
 }
 
+fn glob_arg<'a, 'b>(required: bool) -> Arg<'a, 'b> {
+    Arg::with_name("glob")
+        .value_name("GLOB")
+        .takes_value(true)
+        .required(required)
+        .help(
+            "Glob pattern of files. \
+            PATTERN MUST BE IN \
+            QUOTES TO AVOID SHELL EXPANSION.",
+        )
+}
+
+fn id_arg<'a, 'b>() -> Arg<'a, 'b> {
+    Arg::with_name("id")
+        .value_name("ID")
+        .takes_value(true)
+        .required(true)
+        .validator(is_parsable::<Base64>)
+        .help("Specify the transaction id.")
+}
+
+fn log_dir_arg<'a, 'b>() -> Arg<'a, 'b> {
+    Arg::with_name("log_dir")
+        .long("log-dir")
+        .value_name("LOG_DIR")
+        .takes_value(true)
+        .validator(is_parsable::<PathBuf>)
+        .help(
+            "Directory that status updates will be written to. If not \
+        provided, status updates will not be written.",
+        )
+}
+
+fn max_confirms_arg<'a, 'b>() -> Arg<'a, 'b> {
+    Arg::with_name("max_confirms")
+        .long("min-confirms")
+        .value_name("MAX_CONFIRM")
+        .takes_value(true)
+        .help("Provide maximum number of confirmations to filter statuses by.")
+}
+
+fn statuses_arg<'a, 'b>() -> Arg<'a, 'b> {
+    Arg::with_name("statuses")
+        .long("statuses")
+        .value_name("STATUSES")
+        .takes_value(true)
+        .multiple(true)
+        .possible_values(&["Submitted", "Pending", "Confirmed", "NotFound"])
+        .help("Status codes to filter by. Multiple Ok.")
+}
+
+fn sol_keypair_path_arg<'a, 'b>() -> Arg<'a, 'b> {
+    Arg::with_name("sol_keypair_path")
+        .long("sol-keypair-path")
+        .value_name("SOL_KEYPAIR_PATH")
+        .validator(is_parsable::<PathBuf>)
+        .env("SOL_KEYPAIR_PATH")
+        .required(true)
+        .help(
+            "Path of Solana keypair file to use to pay for transactions. \
+        Will use value from SOL_KEYPAIR_PATH environment variable \
+        if it exists",
+        )
+}
+
+fn tags_arg<'a, 'b>() -> Arg<'a, 'b> {
+    Arg::with_name("tags")
+        .long("tags")
+        .value_name("TAGS")
+        .multiple(true)
+        .takes_value(true)
+        .validator(is_valid_tag)
+        .help(
+            "Specify additional tags for the files as \
+        <NAME>:<VALUE>, separated by spaces. Content-Type tag \
+        will be inferred automatically so not necessary so \
+        include here. Additional tags will be applied
+        to all of the uploaded files.",
+        )
+}
+
 fn get_app() -> App<'static, 'static> {
     let app_matches = App::new(crate_name!())
         .about(crate_description!())
@@ -153,17 +235,15 @@ fn get_app() -> App<'static, 'static> {
                     "Prints the estimated cost of uploading file(s) \
                 matching provided glob.",
                 )
-                .arg(
-                    Arg::with_name("glob")
-                        .value_name("GLOB")
-                        .takes_value(true)
-                        .required(true)
-                        .help(
-                            "Glob pattern of files to be uploaded. \
-                        Run glob-estimate to see how many files will be \
-                        uploaded, total size and cost.",
-                        ),
-                ),
+                .arg(glob_arg(true)),
+        )
+        .subcommand(
+            SubCommand::with_name("estimate-with-sol")
+                .about(
+                    "Prints the estimated cost of uploading file(s) \
+                matching provided glob in SOL.",
+                )
+                .arg(glob_arg(true)),
         )
         .subcommand(
             SubCommand::with_name("wallet-balance")
@@ -183,212 +263,69 @@ fn get_app() -> App<'static, 'static> {
         .subcommand(
             SubCommand::with_name("get-transaction")
                 .about("Gets a transaction from the network and writes to disk as a file.")
-                .arg(
-                    Arg::with_name("id")
-                        .value_name("ID")
-                        .takes_value(true)
-                        .validator(is_parsable::<Base64>)
-                        .help("Specify the id of the transaction to retrieve."),
-                ),
+                .arg(id_arg()),
         )
         .subcommand(
             SubCommand::with_name("upload")
                 .about("Uploads one or more files that match the specified glob.")
-                .arg(
-                    Arg::with_name("glob")
-                        .value_name("GLOB")
-                        .takes_value(true)
-                        .required(true)
-                        .help(
-                            "Glob pattern of files to be uploaded. \
-                        Run get-cost to see how many files will be \
-                        uploaded, total size and cost. PATTERN MUST BE IN \
-                        QUOTES TO AVOID SHELL EXPANSION.",
-                        ),
-                )
-                .arg(
-                    Arg::with_name("log_dir")
-                        .long("log-dir")
-                        .value_name("LOG_DIR")
-                        .takes_value(true)
-                        .validator(is_parsable::<PathBuf>)
-                        .help(
-                            "Directory that status updates will be written to. If not \
-                        provided, status updates will not be written.
-                        ",
-                        ),
-                )
-                .arg(
-                    Arg::with_name("tags")
-                        .long("tags")
-                        .value_name("TAGS")
-                        .multiple(true)
-                        .takes_value(true)
-                        .validator(is_valid_tag)
-                        .help(
-                            "Specify additional tags for the files as \
-                            <NAME>:<VALUE>, separated by spaces. Content-Type tag \
-                            will be inferred automatically so not necessary so \
-                            include here. Additional tags will be applied
-                            to all of the uploaded files.",
-                        ),
-                ),
+                .arg(glob_arg(true))
+                .arg(log_dir_arg())
+                .arg(tags_arg()),
+        )
+        .subcommand(
+            SubCommand::with_name("upload-with-sol")
+                .about("Uploads one or more files that match the specified glob, paying with SOL.")
+                .arg(glob_arg(true))
+                .arg(log_dir_arg())
+                .arg(tags_arg())
+                .arg(sol_keypair_path_arg()),
         )
         .subcommand(
             SubCommand::with_name("get-raw-status")
                 .about("Prints the raw status of a transaction from the network.")
-                .arg(
-                    Arg::with_name("id")
-                        .value_name("ID")
-                        .takes_value(true)
-                        .required(true)
-                        .help("Specify the id of the transaction to print the status for."),
-                ),
+                .arg(id_arg()),
         )
         .subcommand(
             SubCommand::with_name("update-status")
                 .about("Updates statuses stored in `log_dir` from the network.")
-                .arg(
-                    Arg::with_name("glob")
-                        .value_name("GLOB")
-                        .takes_value(true)
-                        .required(true)
-                        .help(
-                            "Glob pattern of files to be uploaded. \
-                        Run glob-estimate to see how many files will be \
-                        uploaded, total size and cost.",
-                        ),
-                )
-                .arg(
-                    Arg::with_name("log_dir")
-                        .long("log-dir")
-                        .value_name("LOG_DIR")
-                        .takes_value(true)
-                        .validator(is_parsable::<PathBuf>)
-                        .help(
-                            "Directory that status updates will be to. If not \
-                        provided, status updates will not be written.
-                        ",
-                        ),
-                ),
+                .arg(glob_arg(true))
+                .arg(log_dir_arg()),
         )
         .subcommand(
             SubCommand::with_name("status-report")
                 .about("Prints a summary of statuses stored in `log_dir`.")
-                .arg(
-                    Arg::with_name("glob")
-                        .value_name("GLOB")
-                        .takes_value(true)
-                        .help(
-                            "Glob pattern of files to be uploaded. \
-                        Run glob-estimate to see how many files will be \
-                        uploaded, total size and cost.",
-                        ),
-                )
-                .arg(
-                    Arg::with_name("log_dir")
-                        .long("log-dir")
-                        .value_name("LOG_DIR")
-                        .takes_value(true)
-                        .required(true)
-                        .validator(is_parsable::<PathBuf>)
-                        .help(
-                            "Directory that status updates will be to. If not \
-                        provided, status updates will not be written.
-                        ",
-                        ),
-                ),
+                .arg(glob_arg(false))
+                .arg(log_dir_arg()),
+        )
+        .subcommand(
+            SubCommand::with_name("generate-manifest")
+                .about("Writes a manifest.json to `log_dir`.")
+                .arg(glob_arg(true))
+                .arg(log_dir_arg()),
         )
         .subcommand(
             SubCommand::with_name("upload-filter")
                 .about("Re-uploads files that meet filter criteria.")
-                .arg(
-                    Arg::with_name("glob")
-                        .value_name("GLOB")
-                        .takes_value(true)
-                        .required(true)
-                        .help("Glob pattern of files to be uploaded."),
-                )
-                .arg(
-                    Arg::with_name("log_dir")
-                        .long("log-dir")
-                        .value_name("LOG_DIR")
-                        .takes_value(true)
-                        .required(true)
-                        .validator(is_parsable::<PathBuf>)
-                        .help(
-                            "Directory that status updates will be to. If not \
-                        provided, status updates will not be written.
-                        ",
-                        ),
-                )
-                .arg(
-                    Arg::with_name("statuses")
-                        .long("statuses")
-                        .value_name("STATUSES")
-                        .takes_value(true)
-                        .multiple(true)
-                        .possible_values(&["Submitted", "Pending", "Confirmed", "NotFound"])
-                        .help("Status codes to filter by. Multiple Ok."),
-                )
-                .arg(
-                    Arg::with_name("max_confirms")
-                        .long("min-confirms")
-                        .value_name("MAX_CONFIRM")
-                        .takes_value(true)
-                        .help("Provide maximum number of confirmations to filter statuses by."),
-                ),
+                .arg(glob_arg(true))
+                .arg(log_dir_arg())
+                .arg(statuses_arg())
+                .arg(max_confirms_arg()),
         )
         .subcommand(
             SubCommand::with_name("list-status")
                 .about("Lists statuses as currently store in `log_dir`.")
                 .help("")
-                .arg(
-                    Arg::with_name("glob")
-                        .value_name("GLOB")
-                        .takes_value(true)
-                        .required(true)
-                        .help(
-                            "Glob pattern of files to be uploaded. \
-                        Run glob-estimate to see how many files will be \
-                        uploaded, total size and cost.",
-                        ),
-                )
-                .arg(
-                    Arg::with_name("log_dir")
-                        .long("log-dir")
-                        .value_name("LOG_DIR")
-                        .takes_value(true)
-                        .required(true)
-                        .validator(is_parsable::<PathBuf>)
-                        .help(
-                            "Directory that status updates will be to. If not \
-                        provided, status updates will not be written.
-                        ",
-                        ),
-                )
-                .arg(
-                    Arg::with_name("statuses")
-                        .long("statuses")
-                        .value_name("STATUSES")
-                        .takes_value(true)
-                        .multiple(true)
-                        .possible_values(&["Submitted", "Pending", "Confirmed", "NotFound"])
-                        .help("Status codes to filter by. Multiple Ok."),
-                )
-                .arg(
-                    Arg::with_name("max_confirms")
-                        .long("min-confirms")
-                        .value_name("MAX_CONFIRM")
-                        .takes_value(true)
-                        .help("Provide maximum number of confirmations to filter statuses by."),
-                ),
+                .arg(glob_arg(true))
+                .arg(log_dir_arg())
+                .arg(statuses_arg())
+                .arg(max_confirms_arg()),
         );
     app_matches
 }
 
 #[tokio::main]
 async fn main() -> CommandResult {
+    env_logger::init();
     let app_matches = get_app().get_matches();
     let keypair_path = app_matches.value_of("keypair_path").unwrap();
     let base_url = app_matches
@@ -407,6 +344,10 @@ async fn main() -> CommandResult {
             let glob_str = sub_arg_matches.value_of("glob").unwrap();
             command_get_cost(&arweave, glob_str).await
         }
+        ("estimate-with-sol", Some(sub_arg_matches)) => {
+            let glob_str = sub_arg_matches.value_of("glob").unwrap();
+            command_get_cost_with_sol(&arweave, glob_str).await
+        }
         ("wallet-balance", Some(sub_arg_matches)) => {
             let wallet_address = sub_arg_matches
                 .value_of("wallet_address")
@@ -425,6 +366,24 @@ async fn main() -> CommandResult {
             let output_format = app_matches.value_of("output_format");
             command_upload(&arweave, glob_str, log_dir, tags, output_format, buffer).await
         }
+        ("upload-with-sol", Some(sub_arg_matches)) => {
+            let glob_str = sub_arg_matches.value_of("glob").unwrap();
+            let log_dir = sub_arg_matches.value_of("log_dir");
+            let tags = get_tags_vec(sub_arg_matches.values_of("tags"));
+            let sol_keypair_path = sub_arg_matches.value_of("sol_keypair_path").unwrap();
+            let buffer = app_matches.value_of("buffer");
+            let output_format = app_matches.value_of("output_format");
+            command_upload_with_sol(
+                &arweave,
+                glob_str,
+                log_dir,
+                tags,
+                sol_keypair_path,
+                output_format,
+                buffer,
+            )
+            .await
+        }
         ("get-raw-status", Some(sub_arg_matches)) => {
             let id = sub_arg_matches.value_of("id").unwrap();
             command_get_raw_status(&arweave, id).await
@@ -463,6 +422,11 @@ async fn main() -> CommandResult {
             let log_dir = sub_arg_matches.value_of("log_dir").unwrap();
             command_status_report(&arweave, glob_str, log_dir).await
         }
+        ("generate-manifest", Some(sub_arg_matches)) => {
+            let glob_str = sub_arg_matches.value_of("glob").unwrap();
+            let log_dir = sub_arg_matches.value_of("log_dir").unwrap();
+            command_generate_manifest(&arweave, glob_str, log_dir).await
+        }
         ("upload-filter", Some(sub_arg_matches)) => {
             let glob_str = sub_arg_matches.value_of("glob").unwrap();
             let log_dir = sub_arg_matches.value_of("log_dir").unwrap();
@@ -506,7 +470,7 @@ async fn command_get_cost(arweave: &Arweave, glob_str: &str) -> CommandResult {
         )
     });
 
-    let (_, usd_per_ar) = arweave.get_price(&1).await?;
+    let (_, usd_per_ar, _) = arweave.get_price(&1).await?;
     let usd_per_bytes = (&cost * &usd_per_ar).to_f32().unwrap() / 1e14_f32;
     println!(
         "The price to upload {} files with {} total bytes is {} {} (${}).",
@@ -515,6 +479,30 @@ async fn command_get_cost(arweave: &Arweave, glob_str: &str) -> CommandResult {
     Ok(())
 }
 
+async fn command_get_cost_with_sol(arweave: &Arweave, glob_str: &str) -> CommandResult {
+    let paths_iter = glob(glob_str)?.filter_map(Result::ok);
+    let (base, incremental) = arweave.get_price_terms().await?;
+
+    let (num, cost, bytes) = paths_iter.fold((0, 0, 0), |(n, c, b), p| {
+        (
+            n + 1,
+            c + {
+                let data_len = p.metadata().unwrap().len();
+                std::cmp::max((base + incremental * (data_len - 1)) / RATE, FLOOR) + 5000
+            },
+            b + p.metadata().unwrap().len(),
+        )
+    });
+
+    let (_, _, usd_per_sol) = arweave.get_price(&1).await?;
+    let usd_per_bytes = (&cost * &usd_per_sol).to_f32().unwrap() / 1e11_f32;
+    println!(
+        "The price to upload {} files with {} total bytes is {} {} (${}).",
+        num, bytes, cost, "lamports", usd_per_bytes
+    );
+    Ok(())
+}
+
 async fn command_get_transaction(arweave: &Arweave, id: &str) -> CommandResult {
     let id = Base64::from_str(id)?;
     let transaction = arweave.get_transaction(&id).await?;
@@ -539,9 +527,11 @@ async fn command_wallet_balance(
         arweave.get_price(&mb)
     );
     let balance = result.0?;
-    let (winstons_per_kb, usd_per_ar) = result.1?;
+    let (winstons_per_kb, usd_per_ar, _) = result.1?;
 
-    let balance_usd = &balance / &WINSTONS_PER_AR * &usd_per_ar;
+    let balance_usd = &balance.to_f32().unwrap() / &WINSTONS_PER_AR.to_f32().unwrap()
+        * &usd_per_ar.to_f32().unwrap()
+        / 100_f32;
 
     let usd_per_kb = (&winstons_per_kb * &usd_per_ar).to_f32().unwrap() / 1e14_f32;
 
@@ -551,7 +541,7 @@ async fn command_wallet_balance(
             units = arweave.units,
             max = &balance / &winstons_per_kb,
             price = &winstons_per_kb,
-            balance_usd = balance_usd.to_f32().unwrap() / 100_f32,
+            balance_usd = balance_usd,
             usd_price = usd_per_kb
     );
     Ok(())
@@ -611,6 +601,67 @@ async fn command_upload(
     Ok(())
 }
 
+async fn command_upload_with_sol(
+    arweave: &Arweave,
+    glob_str: &str,
+    log_dir: Option<&str>,
+    _tags: Option<Vec<Tag>>,
+    sol_keypair_path: &str,
+    output_format: Option<&str>,
+    buffer: Option<&str>,
+) -> CommandResult {
+    let paths_iter = glob(glob_str)?.filter_map(Result::ok);
+    let log_dir = log_dir.map(|s| PathBuf::from(s));
+    let output_format = get_output_format(output_format.unwrap_or(""));
+    let buffer = buffer.map(|b| b.parse::<usize>().unwrap()).unwrap_or(1);
+    let price_terms = arweave.get_price_terms().await?;
+    let solana_url = "https://api.mainnet-beta.solana.com/".parse::<Url>()?;
+    let sol_ar_url = SOL_AR_BASE_URL.parse::<Url>()?.join("sol")?;
+    let from_keypair = keypair::read_keypair_file(sol_keypair_path)?;
+
+    let mut stream = upload_files_with_sol_stream(
+        arweave,
+        paths_iter,
+        log_dir.clone(),
+        None,
+        price_terms,
+        solana_url,
+        sol_ar_url,
+        &from_keypair,
+        buffer,
+    );
+
+    let mut counter = 0;
+    while let Some(result) = stream.next().await {
+        match result {
+            Ok(status) => {
+                if counter == 0 {
+                    if let Some(log_dir) = &log_dir {
+                        println!("Logging statuses to {}", &log_dir.display());
+                    }
+                    println!("{}", Status::header_string(&output_format));
+                }
+                print!("{}", output_format.formatted_string(&status));
+                counter += 1;
+            }
+            Err(e) => println!("{:#?}", e),
+        }
+    }
+
+    if counter == 0 {
+        println!("The pattern \"{}\" didn't match any files.", glob_str);
+    } else {
+        println!(
+            "Uploaded {} files. Run `arloader update-status \"{}\" --log-dir {} to confirm transaction(s).",
+            counter,
+            glob_str,
+            &log_dir.unwrap_or(PathBuf::from("")).display()
+        );
+    }
+
+    Ok(())
+}
+
 async fn command_list_statuses(
     arweave: &Arweave,
     glob_str: &str,
@@ -686,6 +737,21 @@ async fn command_status_report(arweave: &Arweave, glob_str: &str, log_dir: &str)
     Ok(())
 }
 
+async fn command_generate_manifest(
+    arweave: &Arweave,
+    glob_str: &str,
+    log_dir: &str,
+) -> CommandResult {
+    let paths_iter = glob(glob_str)?.filter_map(Result::ok);
+    let log_dir = PathBuf::from(log_dir);
+
+    let _ = arweave.write_manifest(paths_iter, log_dir.clone()).await?;
+
+    println!("manifest.json written to {}", log_dir.display());
+
+    Ok(())
+}
+
 async fn command_upload_filter(
     arweave: &Arweave,
     glob_str: &str,
